@@ -4,13 +4,13 @@
 #include "client.h"
 #include "time_cache.h"
 
-int
+inline int
 open_log_file(const char *path);
 
-void
+inline void
 write_error_log(char *file_name, int line);
 
-int 
+inline int 
 write_access_log(client_t *cli, int log_fd, const char *log_path);
 
 #endif
