@@ -25,9 +25,13 @@
 
 static char *server_name = "127.0.0.1";
 static short server_port = 8000;
-static int listen_sock;  // listen socket
+static int listen_sock = -1;  // listen socket
 
-static int loop_done; // main loop flag
+static uint64_t total_events = 0;
+
+static volatile sig_atomic_t loop_done;
+static volatile sig_atomic_t call_shutdown = 0;
+static volatile sig_atomic_t catch_signal = 0;
 
 picoev_loop* main_loop; //main loop
 
@@ -64,6 +68,7 @@ static PyObject *empty_string = NULL; //""
 /* gunicorn */
 static int spinner = 0;
 static int tempfile_fd = 0;
+static int gtimeout = 0;
 static int ppid = 0;
 
 #define CLIENT_MAXFREELIST 1024
@@ -77,6 +82,9 @@ r_callback(picoev_loop* loop, int fd, int events, void* cb_arg);
 static void
 w_callback(picoev_loop* loop, int fd, int events, void* cb_arg);
 
+static void
+kill_callback(picoev_loop* loop, int fd, int events, void* cb_arg);
+
 static inline void
 resume_wsgi_app(ClientObject *pyclient, picoev_loop* loop);
 
@@ -84,7 +92,7 @@ static inline void
 prepare_call_wsgi(client_t *client);
 
 static inline void
-call_wsgi_app(client_t *client, picoev_loop* loop);
+call_wsgi_app(client_t *client);
 
 static inline int
 check_status_code(client_t *client);
@@ -99,33 +107,48 @@ setsig(int sig, void* handler)
     return sigaction(sig, &context, &ocontext);
 }
 
+static inline void
+kill_server(int timeout)
+{
+    //stop accepting
+    picoev_del(main_loop, listen_sock);
+    total_events--;
+
+    if(timeout > 0){
+        //set timeout
+        picoev_add(main_loop, listen_sock, PICOEV_TIMEOUT, timeout, kill_callback, NULL);
+    }else{
+        picoev_add(main_loop, listen_sock, PICOEV_TIMEOUT, 1, kill_callback, NULL);
+    }
+}
+
 static inline void
 client_t_list_fill(void)
 {
     client_t *client;
-	while (client_numfree < CLIENT_MAXFREELIST) {
+    while (client_numfree < CLIENT_MAXFREELIST) {
         client = (client_t *)PyMem_Malloc(sizeof(client_t));
-		client_free_list[client_numfree++] = client;
-	}
+        client_free_list[client_numfree++] = client;
+    }
 }
 
 static inline void
 client_t_list_clear(void)
 {
-	client_t *op;
+    client_t *op;
 
-	while (client_numfree) {
-		op = client_free_list[--client_numfree];
-		PyMem_Free(op);
-	}
+    while (client_numfree) {
+        op = client_free_list[--client_numfree];
+        PyMem_Free(op);
+    }
 }
 
 static inline client_t*
 alloc_client_t(void)
 {
     client_t *client;
-	if (client_numfree) {
-		client = client_free_list[--client_numfree];
+    if (client_numfree) {
+        client = client_free_list[--client_numfree];
 #ifdef DEBUG
         printf("use pooled client %p\n", client);
 #endif
@@ -142,18 +165,19 @@ alloc_client_t(void)
 static inline void
 dealloc_client(client_t *client)
 {
-	if (client_numfree < CLIENT_MAXFREELIST){
-		client_free_list[client_numfree++] = client;
+    if (client_numfree < CLIENT_MAXFREELIST){
+        client_free_list[client_numfree++] = client;
     }else{
-	    PyMem_Free(client);
+        PyMem_Free(client);
     }
 }
 
 
 static inline client_t *
-new_client_t(int client_fd, char *remote_addr, uint32_t remote_port){
+new_client_t(int client_fd, char *remote_addr, uint32_t remote_port)
+{
     client_t *client;
-    
+
     client = alloc_client_t();
     //client = PyMem_Malloc(sizeof(client_t)); 
     //memset(client, 0, sizeof(client_t));
@@ -203,52 +227,57 @@ clean_cli(client_t *client)
 }
 
 static inline void 
-close_conn(client_t *cli, picoev_loop* loop)
+close_conn(client_t *client)
 {
     client_t *new_client;
-    if(!cli->response_closed){
-        close_response(cli);
+    if(!client->response_closed){
+        close_response(client);
     }
 
-    picoev_del(loop, cli->fd);
-    clean_cli(cli);
+    if(picoev_is_active(main_loop, client->fd)){
+        picoev_del(main_loop, client->fd);
+        total_events--;
+    }
+
+    clean_cli(client);
 
 #ifdef DEBUG
-    printf("start close client:%p fd:%d status_code %d \n", cli, cli->fd, cli->status_code);
-    printf("picoev_del client:%p fd:%d \n", cli, cli->fd);
-    printf("remain http pipeline size :%d \n", cli->request_queue->size);
+    printf("start close client:%p fd:%d status_code %d \n", client, client->fd, client->status_code);
+    printf("picoev_del client:%p fd:%d \n", client, client->fd);
+    printf("remain http pipeline size :%d \n", client->request_queue->size);
 #endif
     
-    if(cli->request_queue->size > 0){
-        if(check_status_code(cli) > 0){
+    if(client->request_queue->size > 0){
+        if(check_status_code(client) > 0){
             //process pipeline 
-            prepare_call_wsgi(cli);
-            call_wsgi_app(cli, loop);
+            prepare_call_wsgi(client);
+            call_wsgi_app(client);
         }
         return ;
     }
 
-    if(cli->http != NULL){
-        PyMem_Free(cli->http);
+    if(client->http != NULL){
+        PyMem_Free(client->http);
     }
 
-    free_request_queue(cli->request_queue);
-    if(!cli->keep_alive){
-        close(cli->fd);
+    free_request_queue(client->request_queue);
+    if(!client->keep_alive){
+        close(client->fd);
 #ifdef DEBUG
-        printf("close client:%p fd:%d status_code %d \n", cli, cli->fd, cli->status_code);
+        printf("close client:%p fd:%d status_code %d \n", client, client->fd, client->status_code);
 #endif
     }else{
-        disable_cork(cli);
-        new_client = new_client_t(cli->fd, cli->remote_addr, cli->remote_port);
+        disable_cork(client);
+        new_client = new_client_t(client->fd, client->remote_addr, client->remote_port);
         new_client->keep_alive = 1;
         init_parser(new_client, server_name, server_port);
         picoev_add(main_loop, new_client->fd, PICOEV_READ, keep_alive_timeout, r_callback, (void *)new_client);
+        total_events++;
     }
     //PyMem_Free(cli);
-    dealloc_client(cli);
+    dealloc_client(client);
 #ifdef DEBUG
-    printf("********************\n\n");
+    printf("****************************\n\n");
 #endif
 
 }
@@ -287,7 +316,7 @@ check_status_code(client_t *client)
 #endif 
         set_current_request(client);
         send_error_page(client);
-        close_conn(client, main_loop);
+        close_conn(client);
         return 0;
     }
     return 1;
@@ -312,9 +341,9 @@ process_resume_wsgi_app(ClientObject *pyclient)
         PyErr_Fetch(&err_type, &err_val, &err_tb);
         PyErr_Clear();
         //set error
-        res = PyGreenlet_Throw(pyclient->greenlet, err_type, err_val, err_tb);
+        res = greenlet_throw(pyclient->greenlet, err_type, err_val, err_tb);
     }else{
-        res = PyGreenlet_Switch(pyclient->greenlet, pyclient->args, pyclient->kwargs);
+        res = greenlet_switch(pyclient->greenlet, pyclient->args, pyclient->kwargs);
     }
     start_response->cli = old_client;
     
@@ -349,7 +378,7 @@ static inline int
 process_wsgi_app(client_t *cli)
 {
     PyObject *args = NULL, *start = NULL, *res = NULL;
-    PyGreenlet *greenlet;
+    PyObject *greenlet;
     ClientObject *pyclient;
     start = create_start_response(cli);
 
@@ -367,12 +396,12 @@ process_wsgi_app(client_t *cli)
 #endif
 
     //new greenlet
-    greenlet = PyGreenlet_New(wsgi_app, NULL);
+    greenlet = greenlet_new(wsgi_app, NULL);
     // set_greenlet
     pyclient->greenlet = greenlet;
     Py_INCREF(pyclient->greenlet);
 
-    res = PyGreenlet_Switch(greenlet, args, NULL);
+    res = greenlet_switch(greenlet, args, NULL);
     //res = PyObject_CallObject(wsgi_app, args);
     Py_DECREF(args);
     Py_DECREF(greenlet);
@@ -409,6 +438,7 @@ switch_wsgi_app(picoev_loop* loop, int fd, PyObject *obj)
     
     //clear event
     picoev_del(loop, fd);
+    total_events--;
     // resume
     resume_wsgi_app(pyclient, loop);
     pyclient->resumed = 0;
@@ -474,6 +504,18 @@ timeout_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
     }
 }
 
+static void
+kill_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
+{
+    if ((events & PICOEV_TIMEOUT) != 0) {
+#ifdef DEBUG
+        printf("force shutdown...\n");
+#endif
+        picoev_del(loop, listen_sock);
+        loop_done = 0;
+    }
+
+}
 
 static void
 w_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
@@ -494,7 +536,7 @@ w_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
 
         //timeout
         client->keep_alive = 0;
-        close_conn(client, loop);
+        close_conn(client);
     
     } else if ((events & PICOEV_WRITE) != 0) {
         ret = process_body(client);
@@ -503,7 +545,7 @@ w_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
 #endif
         if(ret != 0){
             //ok or die
-            close_conn(client, loop);
+            close_conn(client);
         }
     }
 }
@@ -519,7 +561,7 @@ resume_wsgi_app(ClientObject *pyclient, picoev_loop* loop)
             //Internal Server Error
             client->bad_request_code = 500;
             send_error_page(client);
-            close_conn(client, loop);
+            close_conn(client);
             return;
         case 0:
             // suspend
@@ -531,7 +573,7 @@ resume_wsgi_app(ClientObject *pyclient, picoev_loop* loop)
 
     if(client->response_closed){
         //closed
-        close_conn(client, loop);
+        close_conn(client);
         return;
     }
     
@@ -541,7 +583,7 @@ resume_wsgi_app(ClientObject *pyclient, picoev_loop* loop)
             // Internal Server Error
             client->bad_request_code = 500;
             send_error_page(client);
-            close_conn(client, loop);
+            close_conn(client);
             return;
         case 0:
             // continue
@@ -550,17 +592,22 @@ resume_wsgi_app(ClientObject *pyclient, picoev_loop* loop)
             printf("set write callback %d \n", ret);
 #endif
             //clear event
-            picoev_del(loop, client->fd);
+            //
+            if(picoev_is_active(main_loop, client->fd)){
+                picoev_del(main_loop, client->fd);
+                total_events--;
+            }
             picoev_add(loop, client->fd, PICOEV_WRITE, 0, w_callback, (void *)client);
+            total_events++;
             return;
         default:
             // send OK
-            close_conn(client, loop);
+            close_conn(client);
     }
 }
 
 static inline void
-call_wsgi_app(client_t *client, picoev_loop* loop)
+call_wsgi_app(client_t *client)
 {
     int ret;
     ret = process_wsgi_app(client);
@@ -573,7 +620,7 @@ call_wsgi_app(client_t *client, picoev_loop* loop)
             //Internal Server Error
             client->bad_request_code = 500;
             send_error_page(client);
-            close_conn(client, loop);
+            close_conn(client);
             return;
         case 0:
             // suspend
@@ -585,7 +632,7 @@ call_wsgi_app(client_t *client, picoev_loop* loop)
     
     if(client->response_closed){
         //closed
-        close_conn(client, loop);
+        close_conn(client);
         return;
     }
     ret = response_start(client);
@@ -597,7 +644,7 @@ call_wsgi_app(client_t *client, picoev_loop* loop)
             // Internal Server Error
             client->bad_request_code = 500;
             send_error_page(client);
-            close_conn(client, loop);
+            close_conn(client);
             return;
         case 0:
             // continue
@@ -606,12 +653,16 @@ call_wsgi_app(client_t *client, picoev_loop* loop)
             printf("set write callback %d \n", ret);
 #endif
             //clear event
-            picoev_del(loop, client->fd);
-            picoev_add(loop, client->fd, PICOEV_WRITE, 0, w_callback, (void *)client);
+            if(picoev_is_active(main_loop, client->fd)){
+                picoev_del(main_loop, client->fd);
+                total_events--;
+            }
+            picoev_add(main_loop, client->fd, PICOEV_WRITE, 0, w_callback, (void *)client);
+            total_events++;
             return;
         default:
             // send OK
-            close_conn(client, loop);
+            close_conn(client);
     }
     
     
@@ -638,7 +689,7 @@ prepare_call_wsgi(client_t *client)
                     client->keep_alive = 0;
                     client->status_code = 500;
                     send_error_page(client);
-                    close_conn(client, main_loop);
+                    close_conn(client);
                     return;
                 }
             }else{
@@ -646,7 +697,7 @@ prepare_call_wsgi(client_t *client)
                 client->keep_alive = 0;
                 client->status_code = 417;
                 send_error_page(client);
-                close_conn(client, main_loop);
+                close_conn(client);
                 return;
             }
         }
@@ -726,7 +777,7 @@ r_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
             set_bad_request_code(cli, 408);
             finish = 1;
         }else{
-            close_conn(cli, loop);
+            close_conn(cli);
         }
     
     } else if ((events & PICOEV_READ) != 0) {
@@ -749,7 +800,7 @@ r_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
                 }else{
                     cli->status_code = 503;
                     send_error_page(cli);
-                    close_conn(cli, loop);
+                    close_conn(cli);
                     return;
                 }
             case -1: /* error */
@@ -784,7 +835,7 @@ r_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
                                 cli->response_closed = 1;
                             }
                         }
-                        close_conn(cli, loop);
+                        close_conn(cli);
                         return;
                     }
                 }
@@ -844,10 +895,11 @@ r_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
     }
     if(finish == 1){
         picoev_del(loop, cli->fd);
+        total_events--;
         if(check_status_code(cli) > 0){
             //current request ok
             prepare_call_wsgi(cli);
-            call_wsgi_app(cli, loop);
+            call_wsgi_app(cli);
         }
         return;
     }
@@ -873,7 +925,7 @@ accept_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
         client_fd = accept(fd, (struct sockaddr *)&client_addr, &client_len);
         Py_END_ALLOW_THREADS
 
-        if (client_fd != -1) {
+        if (likely(client_fd != -1)) {
 #ifdef DEBUG
             printf("accept fd %d \n", client_fd);
 #endif
@@ -884,12 +936,13 @@ accept_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
             client = new_client_t(client_fd, remote_addr, remote_port);
             init_parser(client, server_name, server_port);
             picoev_add(loop, client_fd, PICOEV_READ, keep_alive_timeout, r_callback, (void *)client);
+            total_events++;
         }else{
             if (errno != EAGAIN && errno != EWOULDBLOCK) {
                 PyErr_SetFromErrno(PyExc_IOError);
                 write_error_log(__FILE__, __LINE__);
                 // die
-                loop_done = 0;
+                kill_server(0);
             }
         }
 
@@ -903,7 +956,6 @@ setup_server_env(void)
     cache_time_init();
     setup_static_env(server_name, server_port);
     setup_start_response();
-    setup_client();
     
     ClientObject_list_fill();
     client_t_list_fill();
@@ -911,8 +963,6 @@ setup_server_env(void)
     header_list_fill();
     buffer_list_fill();
     StringIOObject_list_fill();
-
-    PyGreenlet_Import();
     
     hub_switch_value = Py_BuildValue("(i)", -1);
     client_key = PyString_FromString("meinheld.client");
@@ -1081,26 +1131,38 @@ fast_notify(void)
     spinner = (spinner + 1) % 2;
     fchmod(tempfile_fd, spinner);
     if(ppid != getppid()){
-        loop_done = 0;
+        //shutdown
+        kill_server(gtimeout);
         tempfile_fd = 0;
     }
 }
 
 static PyObject *
-meinheld_listen(PyObject *self, PyObject *args)
+meinheld_listen(PyObject *self, PyObject *args, PyObject *kwds)
 {
-    PyObject *o;
+    PyObject *o = NULL;
     int ret;
+    int sock_fd = -1;
 
-    if (!PyArg_ParseTuple(args, "O:listen", &o))
+    static char *kwlist[] = {"address", "socket_fd", 0};
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|Oi:listen",
+                                     kwlist, &o, &sock_fd)){
         return NULL;
+    }
 
     if(listen_sock > 0){
         PyErr_SetString(PyExc_Exception, "already set listen socket");
         return NULL;
     }
-    
-    if(PyTuple_Check(o)){
+
+    if(o == NULL && sock_fd > 0){
+        listen_sock = sock_fd;
+#ifdef DEBUG
+        printf("use already listened sock fd:%d\n", sock_fd);
+#endif
+        ret = 1;
+    }else if(PyTuple_Check(o)){
         //inet 
         if(!PyArg_ParseTuple(o, "si:listen", &server_name, &server_port))
             return NULL;
@@ -1117,7 +1179,7 @@ meinheld_listen(PyObject *self, PyObject *args)
         listen_sock = -1;
         return NULL;
     }
-
+    
     Py_RETURN_NONE;
 }
 
@@ -1125,16 +1187,19 @@ static void
 sigint_cb(int signum)
 {
 #ifdef DEBUG
-    printf("call SIGINT");
+    printf("call SIGINT\n");
 #endif
-    loop_done = 0;
+    kill_server(0);
+    if(!catch_signal){
+        catch_signal = 1;
+    }
 }
 
 static void 
 sigpipe_cb(int signum)
 {
 #ifdef DEBUG
-    printf("call SIGPIPE");
+    printf("call SIGPIPE\n");
 #endif
 }
 
@@ -1188,27 +1253,42 @@ meinheld_error_log(PyObject *self, PyObject *args)
 
 
 static PyObject *
-meinheld_stop(PyObject *self, PyObject *args)
+meinheld_stop(PyObject *self, PyObject *args, PyObject *kwds)
 {
-    loop_done = 0;
+    int timeout = -1;
+
+    static char *kwlist[] = {"timeout", 0};
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|i:timeout",
+                                     kwlist, &timeout)){
+        return NULL;
+    }
+    kill_server(timeout);
+    Py_RETURN_NONE;
+}
+
+static PyObject *
+meinheld_shutdown(PyObject *self, PyObject *args)
+{
+    call_shutdown = 1;
+    kill_server(0);
     Py_RETURN_NONE;
 }
 
 static PyObject *
 meinheld_run_loop(PyObject *self, PyObject *args)
 {
-    int i = 0;
     //PyObject *app;
     PyObject *watchdog_result;
     if (!PyArg_ParseTuple(args, "O:run", &wsgi_app))
         return NULL; 
-    
+
     if(listen_sock <= 0){
         PyErr_Format(PyExc_TypeError, "not found listen socket");
         return NULL;
-        
+
     }
-    
+
     Py_INCREF(wsgi_app);
     setup_server_env();
 
@@ -1223,13 +1303,11 @@ meinheld_run_loop(PyObject *self, PyObject *args)
     setsig(SIGTERM, sigint_cb);
 
     picoev_add(main_loop, listen_sock, PICOEV_READ, ACCEPT_TIMEOUT_SECS, accept_callback, NULL);
+    total_events++;
     
     /* loop */
-    while (loop_done) {
-        //Py_BEGIN_ALLOW_THREADS
+    while (likely(loop_done == 1 && total_events > 0)) {
         picoev_loop_once(main_loop, 10);
-        //Py_END_ALLOW_THREADS
-        i++;
         // watchdog slow.... skip check
         
         //if(watchdog && i > 1){
@@ -1240,7 +1318,6 @@ meinheld_run_loop(PyObject *self, PyObject *args)
                 PyErr_Clear();
             }
             Py_XDECREF(watchdog_result);
-            i = 0;
         }else if(tempfile_fd){
             fast_notify();
         }
@@ -1248,16 +1325,35 @@ meinheld_run_loop(PyObject *self, PyObject *args)
 
     Py_DECREF(wsgi_app);
     Py_XDECREF(watchdog);
-    
+
     picoev_destroy_loop(main_loop);
     picoev_deinit();
-    
+
     clear_server_env();
 
+    /* if(call_shutdown){ */
+        /* close(listen_sock); */
+        /* listen_sock = 0; */
+        /* if(unix_sock_name){ */
+           /* unlink(unix_sock_name); */
+           /* unix_sock_name = NULL; */
+        /* } */
+        /* call_shutdown = 0; */
+    /* } */
+    close(listen_sock);
+    listen_sock = 0;
     if(unix_sock_name){
-        unlink(unix_sock_name);
+       unlink(unix_sock_name);
+       unix_sock_name = NULL;
+    }
+
+    if(catch_signal){
+        //override
+        PyErr_Clear();
+        PyErr_SetNone(PyExc_KeyboardInterrupt);
+        catch_signal = 0;
+        return NULL;
     }
-    printf("Bye.\n");
     Py_RETURN_NONE;
 }
 
@@ -1386,11 +1482,15 @@ meinheld_set_fastwatchdog(PyObject *self, PyObject *args)
 {
     int _fd;
     int _ppid;
-    if (!PyArg_ParseTuple(args, "ii", &_fd, &_ppid))
+    int timeout;
+
+    if (!PyArg_ParseTuple(args, "iii", &_fd, &_ppid, &timeout)){
         return NULL;
-    
+    }
+
     tempfile_fd = _fd;
     ppid = _ppid;
+    gtimeout = timeout;
     Py_RETURN_NONE;
 }
 
@@ -1443,7 +1543,7 @@ meinheld_suspend_client(PyObject *self, PyObject *args)
     PyObject *temp;
     ClientObject *pyclient;
     client_t *client;
-    PyGreenlet *parent;
+    PyObject *parent;
     int timeout = 0;
 
     if (!PyArg_ParseTuple(args, "O|i:_suspend_client", &temp, &timeout)){
@@ -1476,7 +1576,7 @@ meinheld_suspend_client(PyObject *self, PyObject *args)
     
     if(client && !(pyclient->suspended)){
         pyclient->suspended = 1;
-        parent = PyGreenlet_GET_PARENT(pyclient->greenlet);
+        parent = greenlet_getparent(pyclient->greenlet);
 
         set_so_keepalive(client->fd, 1);
 #ifdef DEBUG
@@ -1484,13 +1584,17 @@ meinheld_suspend_client(PyObject *self, PyObject *args)
         printf("meinheld_suspend_client active ? %d \n", picoev_is_active(main_loop, client->fd));
 #endif
         //clear event
-        picoev_del(main_loop, client->fd);
+        if(picoev_is_active(main_loop, client->fd)){
+            picoev_del(main_loop, client->fd);
+            total_events--;
+        }
         if(timeout > 0){
             picoev_add(main_loop, client->fd, PICOEV_TIMEOUT, timeout, timeout_error_callback, (void *)pyclient);
         }else{
             picoev_add(main_loop, client->fd, PICOEV_TIMEOUT, 300, timeout_callback, (void *)pyclient);
         }
-        return PyGreenlet_Switch(parent, hub_switch_value, NULL);
+        total_events++;
+        return greenlet_switch(parent, hub_switch_value, NULL);
     }else{
         PyErr_SetString(PyExc_Exception, "already suspended");
         return NULL;
@@ -1544,8 +1648,12 @@ meinheld_resume_client(PyObject *self, PyObject *args)
         printf("meinheld_resume_client active ? %d \n", picoev_is_active(main_loop, pyclient->client->fd));
 #endif
         //clear event
-        picoev_del(main_loop, client->fd);
+        if(picoev_is_active(main_loop, client->fd)){
+            picoev_del(main_loop, client->fd);
+            total_events--;
+        }
         picoev_add(main_loop, client->fd, PICOEV_WRITE, 0, resume_callback, (void *)pyclient);
+        total_events++;
     }else{
         PyErr_SetString(PyExc_Exception, "already resumed");
         return NULL;
@@ -1566,6 +1674,7 @@ meinheld_cancel_wait(PyObject *self, PyObject *args)
         return NULL;
     }
     picoev_del(main_loop, fd);
+    total_events--;
     Py_RETURN_NONE;
 }
 
@@ -1585,15 +1694,15 @@ trampoline_switch_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
 static inline PyObject*
 meinheld_trampoline(PyObject *self, PyObject *args, PyObject *kwargs)
 {
-    PyGreenlet *current, *parent;
+    PyObject *current, *parent;
     ClientObject *pyclient;
     int fd, event, timeout = 0;
     PyObject *read = Py_None, *write = Py_None;
 
-	static char *keywords[] = {"fileno", "read", "write", "timeout", NULL};
-	
+    static char *keywords[] = {"fileno", "read", "write", "timeout", NULL};
+    
     if (!PyArg_ParseTupleAndKeywords(args, kwargs, "i|OOi:trampoline", keywords, &fd, &read, &write, &timeout)){
-		return NULL;
+        return NULL;
     }
     
     if(fd < 0){
@@ -1621,14 +1730,18 @@ meinheld_trampoline(PyObject *self, PyObject *args, PyObject *kwargs)
     }
 
     pyclient =(ClientObject *) current_client;
-    
-    picoev_del(main_loop, fd);
+
+    if(picoev_is_active(main_loop, fd)){
+        picoev_del(main_loop, fd);
+        total_events--;
+    }
     picoev_add(main_loop, fd, event, timeout, trampoline_switch_callback, (void *)pyclient);
-   
+    total_events++;
+
     // switch to hub
     current = pyclient->greenlet;
-    parent = PyGreenlet_GET_PARENT(current);
-    return PyGreenlet_Switch(parent, hub_switch_value, NULL);
+    parent = greenlet_getparent(current);
+    return greenlet_switch(parent, hub_switch_value, NULL);
 
 }
 
@@ -1650,7 +1763,7 @@ meinheld_get_ident(PyObject *self, PyObject *args)
 
 
 static PyMethodDef WsMethods[] = {
-    {"listen", meinheld_listen, METH_VARARGS, "set host and port num"},
+    {"listen", (PyCFunction)meinheld_listen, METH_VARARGS | METH_KEYWORDS, "set host and port num or fd"},
     {"access_log", meinheld_access_log, METH_VARARGS, "set access log file path."},
     {"error_log", meinheld_error_log, METH_VARARGS, "set error log file path."},
 
@@ -1670,7 +1783,9 @@ static PyMethodDef WsMethods[] = {
     {"get_picoev_max_fd", meinheld_get_picoev_max_fd, METH_VARARGS, "return picoev max fd size"},
 
     {"set_process_name", meinheld_set_process_name, METH_VARARGS, "set process name"},
-    {"stop", meinheld_stop, METH_VARARGS, "stop main loop"},
+    {"stop", (PyCFunction)meinheld_stop, METH_VARARGS|METH_KEYWORDS, "stop main loop"},
+    {"shutdown", meinheld_shutdown, METH_NOARGS, "stop and close listen socket "},
+
     // support gunicorn 
     {"set_listen_socket", meinheld_set_listen_socket, METH_VARARGS, "set listen_sock"},
     {"set_watchdog", meinheld_set_watchdog, METH_VARARGS, "set watchdog"},
@@ -1709,11 +1824,11 @@ initserver(void)
     }
 
     timeout_error = PyErr_NewException("meinheld.server.timeout",
-					  PyExc_IOError, NULL);
-	if (timeout_error == NULL)
-		return;
-	Py_INCREF(timeout_error);
-	PyModule_AddObject(m, "timeout", timeout_error);
+                      PyExc_IOError, NULL);
+    if (timeout_error == NULL)
+        return;
+    Py_INCREF(timeout_error);
+    PyModule_AddObject(m, "timeout", timeout_error);
 
 #ifdef DEBUG
     printf("client size %d \n", sizeof(client_t));
